///////////////////////////////////////////////////////////////////////////////
// Class C_Loader_Channel                                                    //
// Created by Hell Hibou (2008)                                              //
// Part of Wii Homebrew Launcher                                             //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////


#include <stdlib.h>

#include "kmemory.h"
#include "processor.h"
#include "system.h"
#include "ressources.h"
#include "strconv.h"

#include "loader_Channel.hpp"


#define CHANNEL_TYPE_SYSTEM_BASE			0x00000001L
#define CHANNEL_TYPE_DISK					0x00010000L
#define CHANNEL_TYPE_DOWNLOADED				0x00010001L
#define CHANNEL_TYPE_SYSTEM					0x00010002L
#define CHANNEL_TYPE_GAME					0x00010004L
#define CHANNEL_TYPE_DOWNLOADED_CONTENT		0x00010005L
#define CHANNEL_TYPE_HIDDEN_HOMEBREW		0x00010008L

#define CHANNEL_SYSTEMTYPE_COMMODORE64		0x43
#define CHANNEL_SYSTEMTYPE_NEOGEO			0x45
#define CHANNEL_SYSTEMTYPE_NES				0x46
#define CHANNEL_SYSTEMTYPE_GENERAL			0x48
#define CHANNEL_SYSTEMTYPE_SNES				0x4A
#define CHANNEL_SYSTEMTYPE_MASTERSYSTEM		0x4C
#define CHANNEL_SYSTEMTYPE_MEGADRIVE		0x4D
#define CHANNEL_SYSTEMTYPE_N64				0x4E
#define CHANNEL_SYSTEMTYPE_TURBOGRAFX		0x50
#define CHANNEL_SYSTEMTYPE_TURBOGRAFXCD		0x51
#define CHANNEL_SYSTEMTYPE_WIIDISC			0x52
#define CHANNEL_SYSTEMTYPE_WIIWARE			0x57

#define CHANNEL_REGION_ALL			0x41
#define CHANNEL_REGION_GERMAN		0x44
#define CHANNEL_REGION_USA			0x45
#define CHANNEL_REGION_FRENCH		0x46
#define CHANNEL_REGION_JAPAN		0x4A
#define CHANNEL_REGION_KOREA		0x4B
#define CHANNEL_REGION_PAL			0x4C
#define CHANNEL_REGION_PALWORLD		0x50

#define CHANNEL_TITLE_SIZE	0x54
#define CERTSYS_SIZE		2560

///////////////////////////////////////////////////////////////////////////////

typedef void (*entrypoint) (void);

extern void BeforeExit();

///////////////////////////////////////////////////////////////////////////////

u8 C_Loader_Channel::ValidSU = 0;

///////////////////////////////////////////////////////////////////////////////

int C_Loader_Channel::Compare (const void * a, const void * b)
{
	T_TitleId * TitleA = (T_TitleId*) a;
	T_TitleId * TitleB = (T_TitleId*) b;
	
	// System channels at end
	if ( (TitleA->L3.TypeId == CHANNEL_TYPE_SYSTEM) || (TitleB->L3.TypeId == CHANNEL_TYPE_SYSTEM) )
	{
		if ( (TitleA->L3.TypeId == CHANNEL_TYPE_SYSTEM) && ( TitleB->L3.TypeId == CHANNEL_TYPE_SYSTEM))
		{ if (*(u64*)a > *(u64*)b) { return 1; } else { return -1; } }
		else if (TitleA->L3.TypeId == CHANNEL_TYPE_SYSTEM) { return 1;  } 
		else if (TitleB->L3.TypeId == CHANNEL_TYPE_SYSTEM) { return -1; }
	}
	
	else if ( (TitleA->L3.TypeId == CHANNEL_TYPE_DOWNLOADED) || (TitleB->L3.TypeId == CHANNEL_TYPE_DOWNLOADED) )
	{
		// Wiware first
		if ( (TitleA->L3.SystemId == CHANNEL_SYSTEMTYPE_WIIWARE) && (TitleB->L3.SystemId == CHANNEL_SYSTEMTYPE_WIIWARE) )
		{ if (*(u64*)a > *(u64*)b) { return 1; } else { return -1; } }
		else if (TitleA->L3.SystemId == CHANNEL_SYSTEMTYPE_WIIWARE) { return -1;  } 
		else if (TitleB->L3.SystemId == CHANNEL_SYSTEMTYPE_WIIWARE) { return 1; }
		
		// General channels at end but before System channels
		if ( (TitleA->L3.SystemId == CHANNEL_SYSTEMTYPE_GENERAL) && (TitleB->L3.SystemId == CHANNEL_SYSTEMTYPE_GENERAL) )
		{ if (*(u64*)a > *(u64*)b) { return 1; } else { return -1; } }
		else if (TitleA->L3.SystemId == CHANNEL_SYSTEMTYPE_GENERAL) { return 1;  } 
		else if (TitleB->L3.SystemId == CHANNEL_SYSTEMTYPE_GENERAL) { return -1; }
		
		// Others downloaded channels after WiiWare and before General channels  
		else 
		{ if (*(u64*)a > *(u64*)b) { return 1; } else { return -1; } }
	}
	else
	{
		if (*(u64*)a > *(u64*)b) { return 1; } else { return -1; }
	}
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////

int C_Loader_Channel::Exist ()
{
	if (ValidSU == 0) 
	{
		ISFS_Initialize();
		if (Identify_SU() == NO_ERROR)	{ ValidSU = 1; } else { ValidSU = 2; }
		ISFS_Deinitialize();
	}
	
	if (ValidSU == 1) { return true; }	else 
	{
		return false; 
	}
}

///////////////////////////////////////////////////////////////////////////////

C_Loader_Channel::C_Loader_Channel() 
{
	TextColor	= RGB_GREY; 
	LangId		= CONF_GetLanguage();
	Channel_TitleId = (T_TitleId *) MEM2_malloc (sizeof (T_TitleId) * MAX_CHANNELS);
	ISFS_Initialize();
	Identify_SU();
}

///////////////////////////////////////////////////////////////////////////////

C_Loader_Channel::~C_Loader_Channel() 
{
	MEM2_free(Channel_TitleId);
	ISFS_Deinitialize();
}


///////////////////////////////////////////////////////////////////////////////

void C_Loader_Channel::CleanChannels () 
{
	for (int Boucle = 0; Boucle < MAX_CHANNELS; Boucle++)
	{
		Channel_TitleId [Boucle].TitleId = 0x0LL;
	}
}

///////////////////////////////////////////////////////////////////////////////

void C_Loader_Channel::ReadChannels  ()
{
	u32 		ChannelsCount ATTRIBUTE_ALIGN(32);
	T_TitleId *	TitlesID	= NULL;
	char *		TitleName	= NULL;
	u32 *		Content		= NULL;
	u32			TitleIndex	= 0;
	u32			Counter		= 0;
	u32			TitleSeek;
	s32			PtrFile;
	char		Filename [80];
	T_Image		TempImage;


	
	TempImage.Image = NULL;
	Content = (u32*) MEM2_malloc (32);
	if (Content == NULL) 			{ goto L_C_Loader_Channel_ReadChannels_End; }
	TitleName = (char*) MEM2_malloc (CHANNEL_TITLE_SIZE);
	if (TitleName == NULL) 			{ goto L_C_Loader_Channel_ReadChannels_End; }
	
	if (ES_GetNumTitles(&ChannelsCount) < 0) { goto L_C_Loader_Channel_ReadChannels_End; }
	TitlesID = (T_TitleId*)	 MEM2_malloc(sizeof(u64) * ChannelsCount);
	if (TitlesID == NULL)	{ goto L_C_Loader_Channel_ReadChannels_End; }
	memset (TitlesID, 0x0LL, sizeof(u64) * ChannelsCount);

	if (ES_GetTitles((u64*)TitlesID, ChannelsCount)  < 0) { goto L_C_Loader_Channel_ReadChannels_End; }
	if (ChannelBlocId > ChannelsCount) { ChannelBlocId = 0; }
	qsort (TitlesID, ChannelsCount, sizeof(u64), C_Loader_Channel::Compare);
	
	switch (LangId)
	{
		case CONF_LANG_GERMAN:
		case CONF_LANG_FRENCH:
		case CONF_LANG_SPANISH:
		case CONF_LANG_ITALIAN:
		case CONF_LANG_DUTCH:
			TitleSeek = 0x9C + (CHANNEL_TITLE_SIZE * LangId);
			break;	
		
		default:
			TitleSeek = 0x9C + (CHANNEL_TITLE_SIZE * CONF_LANG_ENGLISH);
			break;
	};

	while (TitleIndex < ChannelsCount)
	{
		if (NbChannels >= MAX_CHANNELS)
		{ 
			ChannelNbNavi |= 2;
			break;
		}
		
		if (TitlesID[TitleIndex].L2.TypeId == CHANNEL_TYPE_SYSTEM_BASE)		{ TitleIndex++; continue; }
		if (TitlesID[TitleIndex].L2.TypeId == CHANNEL_TYPE_DISK)				{ TitleIndex++; continue; }
		if (TitlesID[TitleIndex].L2.TypeId == CHANNEL_TYPE_HIDDEN_HOMEBREW)	{ TitleIndex++; continue; }
		if (TitlesID[TitleIndex].TitleId   == REBOOT_TITLE_ID)					{ TitleIndex++; continue; }
		
		// Recuperation Id du fichier banner.bin
		sprintf(Filename, "/title/%08x/%08x/content/title.tmd", TitlesID[TitleIndex].L2.TypeId , TitlesID[TitleIndex].L2.Id);
		*Content = 0xFFFFFFFF;
		PtrFile = ISFS_Open(Filename, ISFS_OPEN_READ);
		if (PtrFile < 0)							{ TitleIndex++; continue; }
		if (ISFS_Seek (PtrFile,0x1E4,0) < 0)		{ TitleIndex++; continue; };
		if (ISFS_Read (PtrFile, Content, 32) < 0)  { TitleIndex++; continue; };
		ISFS_Close(PtrFile);
		
		// Lecture du titre de la chaine
		memset (TitleName, 0x00, CHANNEL_TITLE_SIZE);
		sprintf(Filename, "/title/%08x/%08x/content/%08x.app", TitlesID[TitleIndex].L2.TypeId , TitlesID[TitleIndex].L2.Id, *Content);
		PtrFile = ISFS_Open(Filename, ISFS_OPEN_READ);
		if (PtrFile < 0) { TitleIndex++; continue; }
		ISFS_Seek(PtrFile,TitleSeek,0);
		ISFS_Read(PtrFile, TitleName, CHANNEL_TITLE_SIZE);
		ISFS_Close(PtrFile);
		
		if (TitleName[1] == 0x00) { TitleIndex++; continue; }
		
		if (Counter < ChannelBlocId) 
		{
			Counter++; 
			TitleIndex++;
		} 
		else
		{
			
			str_convert (Channel[NbChannels].Name, DEFAULT_STRING_PAGE_CODE, TitleName, String_PageCode_UNICODE);
			// sprintf (Channel[NbChannels].Name, "%08x-%08x", TitlesID[TitleIndex].L2.TypeId , TitlesID[TitleIndex].L2.Id);
			
			Channel_TitleId[NbChannels] = TitlesID[TitleIndex];
			Channel[NbChannels].Flags |= INTERFACE_FLAG_SHOWNAME;
			
			if (TitlesID[TitleIndex].L2.TypeId == CHANNEL_TYPE_DOWNLOADED) 
			{
				switch (TitlesID[TitleIndex].L3.SystemId)
				{			
					case CHANNEL_SYSTEMTYPE_COMMODORE64:
					case CHANNEL_SYSTEMTYPE_NEOGEO:
					case CHANNEL_SYSTEMTYPE_NES:
					case CHANNEL_SYSTEMTYPE_SNES:
					case CHANNEL_SYSTEMTYPE_MASTERSYSTEM:
					case CHANNEL_SYSTEMTYPE_MEGADRIVE:
					case CHANNEL_SYSTEMTYPE_N64:
					case CHANNEL_SYSTEMTYPE_TURBOGRAFX:
					case CHANNEL_SYSTEMTYPE_TURBOGRAFXCD:
					case CHANNEL_SYSTEMTYPE_WIIDISC:
					Image_PNG::Decode (&Channel[NbChannels].Icon, pngTitleWiiChannelVC, pngTitleWiiChannelVC_Length);
					break;			
					
					case CHANNEL_SYSTEMTYPE_WIIWARE:
						Image_PNG::Decode (&Channel[NbChannels].Icon, pngTitleWiiChannelWiiWare, pngTitleWiiChannelWiiWare_Length);
						break;			
					
					case CHANNEL_SYSTEMTYPE_GENERAL:
					default:
					Image_PNG::Decode (&Channel[NbChannels].Icon, pngTitleWiiChannelDefault, pngTitleWiiChannelDefault_Length);
						break;
				}
			}
			else { Image_PNG::Decode (&Channel[NbChannels].Icon, pngTitleWiiChannelDefault, pngTitleWiiChannelDefault_Length); }
			
			switch (TitlesID[TitleIndex].L3.Region)
			{
				case CHANNEL_REGION_ALL:
					Image_PNG::Decode (&TempImage, pngTitleZoneAll, pngTitleZoneAll_Length);
					break;
				
				case CHANNEL_REGION_GERMAN:
					Image_PNG::Decode (&TempImage, pngTitleZoneGe, pngTitleZoneGe_Length);
					break;
				
				case CHANNEL_REGION_USA:
					Image_PNG::Decode (&TempImage, pngTitleZoneUs, pngTitleZoneUs_Length);
					break;
				
				case CHANNEL_REGION_FRENCH:
					Image_PNG::Decode (&TempImage, pngTitleZoneFr, pngTitleZoneFr_Length);
					break;
				
				case CHANNEL_REGION_JAPAN:
					Image_PNG::Decode (&TempImage, pngTitleZoneJp, pngTitleZoneJp_Length);
					break;
				
				case CHANNEL_REGION_KOREA:
					Image_PNG::Decode (&TempImage, pngTitleZoneKr, pngTitleZoneKr_Length);
					break;
					
				case CHANNEL_REGION_PALWORLD:
				case CHANNEL_REGION_PAL:
					Image_PNG::Decode (&TempImage, pngTitleZonePal, pngTitleZonePal_Length);
					break;
			}
			
			if (TempImage.Image != NULL)
			{
				C_GFX::FusionImages (&Channel[NbChannels].Icon, 0, Channel[NbChannels].Icon, 0, TempImage, 0); 
				MEM2_free(TempImage.Image);
				TempImage.Image = NULL;
			}
			
			C_GFX::FusionImages (&Channel[NbChannels].Icon, 0, Channel[NbChannels].Icon, 0, ImageRessource[ImageRessource_WiiBar], 0); 
			C_GFX::SetImageAlphaMask (Channel[NbChannels].Icon, ImageRessource[ImageRessource_ThumbSprite]);						
			NbChannels++;
			TitleIndex++; 
		}
	}
	
L_C_Loader_Channel_ReadChannels_End:
	if (Content   != NULL) { MEM2_free (Content);		}
	if (TitleName != NULL) { MEM2_free (TitleName);	}
	if (TitlesID  != NULL) { MEM2_free (TitlesID);		}
}

///////////////////////////////////////////////////////////////////////////////

int C_Loader_Channel::Identify_SU()
{
	static const char StructIssuer [] = "Root-CA%08lx-%s%08lx";
	T_SU_TIK * Tik	= NULL;
	T_SU_TMD * Tmd	= NULL;
	u8 * CertSys	= NULL;
	u32 KeyID, PtrFile;
	int RetVal = NO_ERROR;	

	
	// Recuperation du certificat 'Super User'
	CertSys = (u8*) MEM2_malloc (CERTSYS_SIZE);		// memalign(32, CERTSYS_SIZE);
	if(CertSys == NULL)				{ RetVal = ERROR_OUT_OF_MEMORY; goto L_Identify_SU_End; }
	PtrFile = ISFS_Open ("/sys/cert.sys", ISFS_OPEN_READ);
	if (PtrFile < 0)				{ RetVal = ERROR_SYSTEM; goto L_Identify_SU_End; }
	ISFS_Read (PtrFile, CertSys, CERTSYS_SIZE);
	ISFS_Close(PtrFile);


	// Creation du ticket 'Super User'
	Tik = (T_SU_TIK*) MEM2_malloc (sizeof(T_SU_TIK));	// memalign(32, sizeof(T_SU_TIK));
	if (Tik == NULL) { RetVal = ERROR_OUT_OF_MEMORY; goto L_Identify_SU_End; }
	memset (Tik, 0x00, sizeof (T_SU_TIK));
	Tik->TIK.Sign.type = ES_SIG_RSA2048;
	sprintf (Tik->TIK.Ticket.issuer, StructIssuer, 1, "XS", 3);	
	Tik->TIK.Ticket.ticketid	= 0x000038A45236EE5FLL;
	Tik->TIK.Ticket.titleid		= 0x0000000100000000LL;
	Tik->Data [0x1F0]			= 0x12;
	Tik->Data [0x1F1]			= 0x02;
	for (int Boucle = 0x222; Boucle < 0x242; Boucle++) { Tik->Data[Boucle] = 0xFF; }
	
	
	// Creation du TMD 'Super User'
	Tmd = (T_SU_TMD*) MEM2_malloc (sizeof(T_SU_TMD));	// memalign(32, sizeof(T_SU_TMD));
	if (Tik == NULL) { RetVal = ERROR_OUT_OF_MEMORY; goto L_Identify_SU_End; }
	memset (Tmd, 0x00, sizeof (T_SU_TMD));
	Tmd->Sign.type = ES_SIG_RSA2048;
	sprintf (Tmd->TMD.issuer, StructIssuer, 1, "CP", 4);	
	Tmd->TMD.title_id		= 0x0000000100000000LL;
	Tmd->TMD.zero			= 0x1700;
	Tmd->TMD.num_contents	= 1;


	// Passage en mode 'Super User'
	RetVal = ES_Identify ((signed_blob*)CertSys, CERTSYS_SIZE, (signed_blob*) Tmd, sizeof(T_SU_TMD), (signed_blob*) Tik, sizeof(T_SU_TIK), &KeyID);

L_Identify_SU_End:
	if (Tmd != NULL)		{ MEM2_free (Tmd);		}
	if (Tik != NULL)		{ MEM2_free (Tik);		}
	if (CertSys != NULL)	{ MEM2_free(CertSys);	}	
	return RetVal;
}

///////////////////////////////////////////////////////////////////////////////

int C_Loader_Channel::Run (int ChannelId)
{
	static tikview * Tickets;
	u32 TicketsCount;
	int RetVal;

	RetVal = ES_GetNumTicketViews (Channel_TitleId[ChannelId].TitleId, &TicketsCount);
	if (RetVal < 0)			{ return ERROR_BAD_DATA; }
	if (!TicketsCount)		{ return ERROR_UNKNOW; }
	Tickets = (tikview*) MEM2_malloc (sizeof(tikview) * TicketsCount);
	if (Tickets == NULL)	{ return ERROR_OUT_OF_MEMORY; }

	RetVal = ES_GetTicketViews(Channel_TitleId[ChannelId].TitleId, Tickets, TicketsCount);

	switch (RetVal)
	{
		case ES_EINVAL:		MEM2_free (Tickets); return ES_EINVAL;
		case ES_ENOMEM:		MEM2_free (Tickets); return ERROR_OUT_OF_MEMORY;
		case ES_ENOTINIT:	MEM2_free (Tickets); return ES_ENOTINIT;
		case ES_EALIGN:		MEM2_free (Tickets); return ERROR_BAD_DATA;
	}
	
	if (RetVal < 0) { MEM2_free (Tickets); return ERROR_UNKNOW; }
	
	BeforeExit();
	RetVal = ES_LaunchTitle(Channel_TitleId[ChannelId].TitleId, &Tickets[0]);
	Reload ();
	return ERROR_SYSTEM;
}

///////////////////////////////////////////////////////////////////////////////
